/*****************************************************************************
                        Copyright (c) Jens H. Holm
******************************************************************************/
/*
The solution database is used to store solutions and tables during a session with Pro/3.
chain "H" (one record per sol.set)
      domain STERM format: cmp("H",
                              [int(SOLREF),
                              int(NSOL),       #solutions
                              int(TRACED),    1=sol.set has been traced  0=else
                              int(NCOMPS),    #components
                              QUERY,           original query which set is a response to
                              int(ALTERED)])  1=set has been altered  0=else
      indexes: H1:SOLREF (un)
       
chain "V" (one record per sol. set)
      domain STERM format: cmp("V",
                              [int(SOLREF),
                               int(COMPNO),
                               str(VAR)]
      indexes: V1:SOLREF|COMPNO (un)
               V2:SOLREF|VAR    (un)                            
               V3:SOLREF        (nu)
	
chain "Snn.." (one record per solution in sol.set nn..)
      domain STERM format: cmp("S",[int(SolNo)|VALUES])
                           note that SolNo is 0 while solutions are being inserted and
                           thereafter set to 1,..,N (after call on solSetEnumerate)
      indexes: Snn..:SOLNO  (un)  note that there is one index table per chain
      
1/ queries on unary relations:
interpretation	solution	components
--------------	---------------	---------------------------------------------------------------------
attributes	entity		the entity's attributes:
(same entity			1/1:OUTRIGHTcontract_no=str("O-1")
type)				1/2:OUTRIGHTtrade_dt=int(9054)
entity		entity		only one component:
(different			1/1:X=cmp("spot",[str("S-1"),int(9054),str("direct"),str("Dresdner,
entity				HK"),fix(1000000),str("USD"),fix(1550000),str("DEM"),str("Deutsche, Frankfurt"),
types)				str("NDEMDB-101"),str("Citibank, NY"),str("FUSDCB-301")])  
2/ queries on binary relations:
each solution represents the solution for all arguments.
each argument is represented as in either of the two case above (the representation for the two
arguments does not need to be the same).
*/
implement sol
open core,termDomain

class facts
solset : (integer SolSet,string_list RelComps) nondeterm. % RelComps : only those are stored in the set (ascending order)
	
constants
className = "com/rudstrand/dba/sol/sol".
classVersion = "1.0".

class predicates
solMess : (integer,integer,integer) determ (i,i,i).

clauses
classInfo(className, classVersion).
% -----------------------------------------------------------------------------------------------
% dropSolSet/addSolSet
% -----------------------------------------------------------------------------------------------
dropSolSet(SolRef) :- retractall(solset(SolRef,_)),!.
addSolSet(SolRef,Comps) :- assert(solset(SolRef,Comps)),!.
exSolSet(SolRef,Comps) :- solset(SolRef,Comps),!.
% -----------------------------------------------------------------------------------------------
% forallsol
% -----------------------------------------------------------------------------------------------
% returns all solutions non-deterministically.
% can handle gaps in the solution numbers (ie. caused by retractions)
% fails if no more solutions
%
forallsol(SolRef,SolNo) :-
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[_,int(N)|_]),
	Ix=string::concat("S",Key),
	SN=sdb::getIx(Ix),
	!,
	SolNo=programControl::sequence_nd(1,N,1),
	   SSolNo=toString(SolNo),
	   SDB:key_search(SN,SSolNo,_). % to check that solution exists
% -----------------------------------------------------------------------------------------------
% forallsolsorted
% -----------------------------------------------------------------------------------------------
forallsolsorted(SolRef,First,LogNo,PhysNo) :-
	sdb::tabName(SolRef,1,TabName),
	Ix=string::concat("X",TabName),
	XI=sdb::getIx(Ix),
	solSetSize(SolRef,N),
	!,
	LogNo=programControl::sequence_nd(First,N,1),
	  stb::stbGetRowL_P_F(XI,LogNo,PhysNo).
% --------------------------------------------------------------------------
% solCompNo
% --------------------------------------------------------------------------
% returns solution component given a component no
%
solCompNo(SolRef,SolNo,CompNo,Val) :-
	SSolRef=toString(SolRef),
	Ix=string::concat("S",SSolRef),
	SN=sdb::getIx(Ix),
	Key=toString(SolNo),
	SDB=sdb::get(),
	SDB:key_search(SN,Key,Ref),
	SDB:ref_term(Ref,SRec),
	SRec=cmp(_,[_|Vals]),
	set::setGet(Vals,CompNo,Val),
	!
;
	dlgIF::feil("solCompNo",[],[SolRef,SolNo,CompNo],nill),!,fail.	
% --------------------------------------------------------------------------
% solCompNoReplace/VarReplace
% --------------------------------------------------------------------------
% replaces the given solution component
%
solCompNoReplace(SolRef,SolNo,CompNo,NewVal) :-
	S1=toString(SolRef),
	Ix=string::concat("S",S1),
	SN=sdb::getIx(Ix),
	Key=toString(SolNo),
	SDB=sdb::get(),
	SDB:key_search(SN,Key,Ref),
	SDB:ref_term(Ref,SRec),
	SRec=cmp(_,[A|Vals]),
	set::setReplace(Vals,CompNo,NewVal,NewVals),
	NRec=cmp("S",[A|NewVals]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("solCompNoReplace",[],[SolRef,SolNo,CompNo],NewVal),!,fail.	
solVarReplace(SolRef,P,Comp,Val) :-
	solVarNo(SolRef,Comp,CompNo),
	solCompNoReplace(SolRef,P,CompNo,Val),
	!.
% --------------------------------------------------------------------------
% solComps/T
% --------------------------------------------------------------------------
class predicates
solCompsX : (integer,integer,integer,string_list) determ (i,i,i,o).
solCompsTX : (integer,integer,integer,sterml) determ (i,i,i,o).
clauses
% returns list of solution components on string format
%
solComps(SolRef,NComps,Comps) :-
	solSetHeader(SolRef,_,_,NComps,_,_),
	solCompsX(1,NComps,SolRef,Comps),
	!
;
	dlgIF::feil("solComps",[],[SolRef],nill),!,fail.	
% returns list of solution components on term format
%
solCompsT(SolRef,NComps,Comps) :-
	solSetHeader(SolRef,_,_,NComps,_,_),
	solCompsTX(1,NComps,SolRef,Comps),
	!
;
	dlgIF::feil("solCompsT",[],[SolRef],nill),!,fail.	
solCompsX(N,M,_,[]) :-
	N>M,
	!.
solCompsX(N,M,SolRef,[Comp|Comps]) :-
	V1=sdb::getIx("V1"),
	SSolRef=toString(SolRef),
	S2=toString(N),
	sx::concat([SSolRef,"|",S2],Key),
	SDB=sdb::get(),
	SDB:key_search(V1,Key,Ref),
	SDB:ref_term(Ref,VRec),
	VRec=cmp(_,[_,_,str(Comp)]),
	Next=N+1,
	solCompsX(Next,M,SolRef,Comps),
	!
;
	dlgIF::feil("solCompsX",[],[N,M,SolRef],nill),!,fail.	
solCompsTX(N,M,_,[]) :-
	N>M,
	!.
solCompsTX(N,M,SolRef,[Comp|Comps]) :-
	V1=sdb::getIx("V1"),
	SSolRef=toString(SolRef),
	S2=toString(N),
	sx::concat([SSolRef,"|",S2],Key),
	SDB=sdb::get(),
	SDB:key_search(V1,Key,Ref),
	SDB:ref_term(Ref,VRec),
	VRec=cmp(_,[_,_,Comp]),
	Next=N+1,
	solCompsTX(Next,M,SolRef,Comps),
	!
;
	dlgIF::feil("solCompsTX",[],[N,M,SolRef],nill),!,fail.	
% --------------------------------------------------------------------------
% solCompVar
% --------------------------------------------------------------------------
% returns solution component value given a variable
%
solCompVar(SolRef,SolNo,Var,Val) :-
	solVarNo(SolRef,Var,CompNo),
	solCompNo(SolRef,SolNo,CompNo,Val),
	!.
% --------------------------------------------------------------------------
% solCopyRec
% --------------------------------------------------------------------------
% copies given solset record from solref to solef2
%
solCopyRec(SolRef,SolNo,SolRef2) :-
	solGet(SolRef,SolNo,Vals),
	solCompsT(SolRef,_,Comps),
	solInsVals(SolRef2,Comps,Vals),
	!.
% --------------------------------------------------------------------------
% solDel
% --------------------------------------------------------------------------
% deletes a solution from the solution set - no re-enumeration is carried out
% or change to header record
%
solDel(SolRef,SolNo) :-
	SSolRef=toString(SolRef),
	Ix=string::concat("S",SSolRef),
	SN=sdb::getIx(Ix),
	Key=toString(SolNo),
	SDB=sdb::get(),
	SDB:key_search(SN,Key,Ref),
	SDB:term_delete(Ix,Ref),
	SDB:key_delete(SN,Key,Ref),
	!
;
	dlgIF::feil("solDel",[],[SolRef,SolNo],nill),!.	
% --------------------------------------------------------------------------
% solFirstL
% --------------------------------------------------------------------------
% finds the first logical row in a sorted table with possible gaps.
% fails if there are no rows.
%
solFirstL(SolRef,First) :- forallsolsorted(SolRef,1,First,_),!.
% --------------------------------------------------------------------------
% solGet
% --------------------------------------------------------------------------
class predicates
solGetX : (integer,integer,integer,integer,sterml) determ (i,i,i,i,o).
clauses
% returns a complete value list corresponding to a solution
%
solGet(SolRef,SolNo,Vals) :-
	solSetHeader(SolRef,_,_,NComps,_,_),
	solGetX(SolRef,SolNo,1,NComps,Vals),
	!.
solGetX(_,_,N,Last,[]) :-
	N>Last,
	!.
solGetX(SolRef,SolNo,N,Last,[Val|Vals]) :-
	solCompNo(SolRef,SolNo,N,Val),
	Next=N+1,
	solGetX(SolRef,SolNo,Next,Last,Vals),
	!.
% --------------------------------------------------------------------------
% solIns; solInsVals
% --------------------------------------------------------------------------
class predicates
solInsCheckFirst : (string,integer) procedure (i,o).
solInsCheckFirstVals : (string,integer) procedure (i,o).
solInsX : (integer,integer,string,sterml,sterm) determ (i,i,i,i,i).
solInsSRec : (integer,string,sterm) determ (i,i,i).
solRecX : (globalRefDomains::env,sterml,sterml) determ (i,o,o).
solRecPickComps : (string_list,globalRefDomains::env,sterml,sterml) determ (i,i,o,o).
solPick : (string,globalRefDomains::env,sterm,sterm) determ (i,i,o,o).
solRec : (integer,globalRefDomains::env,sterml,sterml) determ (i,i,o,o).
clauses
% inserts a new solution into sdb
%
solIns(SolRef,Env) :-
	SSolRef=toString(SolRef),
	Chain=string::concat("S",SSolRef),
	solInsCheckFirst(Chain,First),
	solRec(SolRef,Env,Vars,Vals),
	% Vars: [str(DE_Var1),str(DE_Var2),...]
	% Vals: [atom(..),int(..),str(..),...]
	solInsX(First,SolRef,Chain,Vars,cmp("S",[int(0)|Vals])),
	!,free(Env),!. % (this call is required to generate more than one solution)
solInsVals(SolRef,Comps,Vals) :-
	SSolRef=toString(SolRef),
	Chain=string::concat("S",SSolRef),
	solInsCheckFirstVals(Chain,First),
	solInsX(First,SolRef,Chain,Comps,cmp("S",[int(0)|Vals])),
	!.
% parameters:
% 1	in	sol.ref.
% 2	in	chain
% 3	in	environment
% 4	out	1/-1 first solution
%
solInsCheckFirst(Chain,First) :-
    SDB=sdb::get(),
	SDB:chain_first(Chain,_), % @@ consider a faster approach to determine if first solution (24.8.99)
	First=-1,
	!
;
	First=1,
	!
;
	First=1,
	!.	
solInsCheckFirstVals(Chain,First) :-
    SDB=sdb::get(),
	SDB:chain_first(Chain,_), % @@ consider a faster approach to determine if first solution (24.8.99)
	First=-1,
	!
;
	First=1,
	!.
% parameters:
% 1	in	1/-1 first solution insertion
% 2 	in	sol.ref
% 3 	in	chain
% 4	in	list of solution record components 	(variables)
% 5	in	do.					(values)
%
solInsX(-1,SolRef,Chain,_,Vals) :-
	solInsSRec(SolRef,Chain,Vals),
	!.
solInsX(1,SolRef,Chain,Vars,Vals) :-
	solInsVRec(SolRef,Vars,1),
			S1=toString(SolRef),
			sx::concat(["SDB/V-insert (Set=",S1,")"],S2),
			tt::tt_terms(4,S2,Vars),
	solInsSRec(SolRef,Chain,Vals),
	!.
% inserts the solution record itself 
% parameters:
% 1	in	sol.ref.
% 2	in	chain
% 3	in	solution record component values
%
solInsSRec(SolRef,Chain,Vals) :-
	not(solset(SolRef,_)), 
	% this is not a condition-set where only required components are inserted i.e. apparent duplicates cannot be suppressed
	SDB=sdb::get(),
	SDB:chain_terms(Chain,ValsX,_),
	Vals=ValsX,
			RL=globalVars::get_i("piercl"),
			solMess(-1,SolRef,RL),
	% exists already (duplicate)		
	!
;
	tt::getLevel()<4,
	SDB=sdb::get(),
	SDB:chain_insertz(Chain,Vals,_),
			RL=globalVars::get_i("piercl"),
			solMess(1,SolRef,RL),
	!
;
	SDB=sdb::get(),
	SDB:chain_insertz(Chain,Vals,_),
			RL=globalVars::get_i("piercl"),
			solMess(1,SolRef,RL),
			S1=toString(SolRef),
			sx::concat(["SDB/S-insert (Set=",S1,")"],S2),
			tt::tt_term(4,S2,Vals),
	!.
solInsVRec(SolRef,[],Next) :- 
	% update H-record with no. of components
	N=Next-1,
	solSetNCompsSet(SolRef,N),
	!.	
solInsVRec(SolRef,[V|Vs],CompNo) :-
	VRec=cmp("V",[int(SolRef),int(CompNo),V]),
	SDB=sdb::get(),
	SDB:chain_inserta("V",VRec,Ref),
	%
    V1=sdb::getIx("V1"),
	SSolRef=toString(SolRef),
	S2=toString(CompNo),
	sx::concat([SSolRef,"|",S2],Key1),
	SDB:key_insert(V1,Key1,Ref),
	%
	V2=sdb::getIx("V2"),
	V=str(S3),
	sx::concat([SSolRef,"|",S3],Key2),
	SDB:key_insert(V2,Key2,Ref),
	%
	V3=sdb::getIx("V3"),
	SDB:key_insert(V3,SSolRef,Ref),
	%
	Next=CompNo+1,
	solInsVRec(SolRef,Vs,Next),
	!.
solInsVRec(SolRef,Vs,CompNo) :- 
	dlgIF::feil("solInsVRec",[],[SolRef,CompNo],cmp("",Vs)),!,fail.	
% converts from PIE environment to sdb formats.
% note that only required solution components are returned (on the basis of the required-flags)
% parameters:
% 1	in	sol.ref.
% 2	in	environment
% 3	out	solution components (variables)
% 4	out	solution components (values)
%
solRec(SolRef,Env,Vars,Vals) :-
	solset(SolRef,Comps),
	!,solRecPickComps(Comps,Env,Vars,Vals),
	!
;
	solRecX(Env,Vars,Vals),
	!.	
solRecX([Es],[],[]) :- 
	free(Es),
	!.
solRecX([globalRefDomains::e(Var,Val)|Es],[SVar|SVars],[SVal|SVals]) :-
	S=Var,
	SVar=str(S),
	conversion::term_sterm(Val,SVal), 
	solRecX(Es,SVars,SVals),
	!.
solRecPickComps([],_,[],[]) :-
	!.
solRecPickComps([Comp|Comps],Env,[SVar|SVars],[SVal|SVals]) :-
	solPick(Comp,Env,SVar,SVal),
	solRecPickComps(Comps,Env,SVars,SVals),
	!.
solRecPickComps(Comps,_,_,_) :-
	dlgIF::feil("solRecPick",["Can't find solution component"|Comps],[],nill),!,fail.
solPick(Comp,[globalRefDomains::e(Var,Val)|Es],SVar,SVal) :-
	S=Var,
	S=Comp,
	SVar=str(S),
	conversion::term_sterm(Val,SVal), 
	!
;
	solPick(Comp,Es,SVar,SVal),
	!.
% --------------------------------------------------------------------------
% solMess
% --------------------------------------------------------------------------
class predicates
solMessCnt : (integer,integer) procedure (i,o).
clauses
solMess(1,SolRef,RL) :-
	RL=1,
	solMessCnt(RL,N),
	Sn=toString(N),
	Mb=string::concat("#",Sn),
	globalVars::set_s("ms2",Mb),
	Ma=globalVars::get_s("ms1"),
	sx::concat([Ma," ",Mb],Mess),
	dlgIF::msgStatus(Mess),
	!
;
	RL>1,	
	solMessCnt(RL,N),
	Sn=toString(N),
	SRL=toString(RL),
	SSolRef=toString(SolRef),
	Ma=globalVars::get_s("ms1"),
	Mb=globalVars::get_s("ms2"),
	sx::concat([Ma," ",Mb,"(#",Sn," in set ",SSolRef," at level ",SRL,")"],Mess),
	dlgIF::msgStatus(Mess),
	!.
solMess(-1,SolRef,RL) :-
	RL=1,
	solMessCnt(RL,N),
	Sn=toString(N),
	Mb=string::concat("#",Sn),
	globalVars::set_s("ms2",Mb),
	Ma=globalVars::get_s("ms1"),
	sx::concat([Ma," ",Mb," dup"],Mess),
	dlgIF::msgStatus(Mess),
	!
;
	RL>1,	
	solMessCnt(RL,N),
	Sn=toString(N),
	SRL=toString(RL),
	SSolRef=toString(SolRef),
	Ma=globalVars::get_s("ms1"),
	Mb=globalVars::get_s("ms2"),
	sx::concat([Ma," ",Mb,"(#",Sn," in set ",SSolRef," at level ",SRL,") dup"],Mess),
	dlgIF::msgStatus(Mess),
	!.
% counts inserted solution records at each recursion level (2 to 5)
%
solMessCnt(1,N) :- globalVars::incr("solcnt0",_,N),!.
solMessCnt(2,N) :- globalVars::incr("solcnt1",_,N),!.
solMessCnt(3,N) :- globalVars::incr("solcnt2",_,N),!.
solMessCnt(4,N) :- globalVars::incr("solcnt3",_,N),!.
solMessCnt(5,N) :- globalVars::incr("solcnt4",_,N),!.
solMessCnt(_,0) :- !.
% --------------------------------------------------------------------------
% solNoVar
% --------------------------------------------------------------------------
solNoVar(SolRef,No,Var) :-
	V1=sdb::getIx("V1"),
	SSolRef=toString(SolRef),
	S2=toString(No),
	sx::concat([SSolRef,"|",S2],Key),
	SDB=sdb::get(),
	SDB:key_search(V1,Key,Ref),
	SDB:ref_term(Ref,VRec),
	VRec=cmp(_,[_,_,str(Var)]),
	!
;
	dlgIF::feil("solNoVar",[],[SolRef,No],nill),!,fail.	
% --------------------------------------------------------------------------
% solSetDel
% --------------------------------------------------------------------------
class predicates
solSetDelCh : (string) procedure.
solSetDelX : (string,integer) determ.
solSetDelV : (string,chainDB::bt_selector) procedure.
clauses

/* 27.5.5
solSetDelCond(SolRef) :-
	globalVars::get_i("ssreus")=1,
	solSetDel(SolRef),
	!
;
	!.
*/
		
% deletes a solution set
%
solSetDel(SolRef) :-
	Ix1=sdb::getIx("H1"),
	Key1=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(Ix1,Key1,Ref1),
	SDB:key_delete(Ix1,Key1,Ref1),
	SDB:ref_term(Ref1,HRec),
	SDB:term_delete("H",Ref1),
	HRec=cmp(_,[_,int(NSol)|_]),
	solSetDelX(Key1,NSol),
	stb::stbDeleteSolAll(SolRef),
	retractall(solset(SolRef,_)),
	!
;
	dlgIF::feil("solSetDel",[],[SolRef],nill),!,fail.	
	
solSetDelX(_,0) :- 
	!.
solSetDelX(SSolRef,_) :-	
	V3=sdb::getIx("V3"),
	solSetDelV(SSolRef,V3),
	Chain=string::concat("S",SSolRef),
	SN=sdb::getIx(Chain),
	SDB=sdb::get(),
	SDB:bt_close(SN),
	SDB:bt_delete(Chain),
	sdb::dropIx(Chain,SN),
	solSetDelCh(Chain),
			tt::tt_s(2,"deleted solution set",SSolRef),!.
	
solSetDelCh(Chain) :-
	SDB=sdb::get(),
	SDB:chain_first(Chain,_), 
	SDB:chain_delete(Chain),
	!
;
	% chain is empty (non-existing)
	!.	
solSetDelV(SSolRef,V3) :-
	SDB=sdb::get(),
	SDB:key_search(V3,SSolRef,Ref),
	SDB:key_delete(V3,SSolRef,Ref),
	SDB:ref_term(Ref,VRec),
	VRec=cmp(_,[_,int(CompNo),str(Var)]),
	SDB:term_delete("V",Ref),
	V1=sdb::getIx("V1"),
	S1=toString(CompNo),
	sx::concat([SSolRef,"|",S1],Key1),
	SDB:key_search(V1,Key1,Ref1),
	SDB:key_delete(V1,Key1,Ref1),
	V2=sdb::getIx("V2"),
	sx::concat([SSolRef,"|",Var],Key2),
	SDB:key_search(V2,Key2,Ref2),
	SDB:key_delete(V2,Key2,Ref2),
	solSetDelV(SSolRef,V3),
	!
;
	!.	
% --------------------------------------------------------------------------
% solSetDelAll
% --------------------------------------------------------------------------
% deletes all solution sets (and tables)
%
solSetDelAll () :-
	SDB=sdb::get(),
	SDB:chain_terms("H",HRec,_),
	HRec=cmp(_,[int(SolRef)|_]),
	solSetDel(SolRef),
	fail
;
	!.	
% --------------------------------------------------------------------------
% solSetDelAltered
% --------------------------------------------------------------------------
solSetDelAltered () :-
	SDB=sdb::get(),
	SDB:chain_terms("H",HRec,_),
	HRec=cmp(_,[int(SolRef),_,_,_,_,int(1)]),
	solSetDel(SolRef),
	fail
;
	!.	
% --------------------------------------------------------------------------
% solSetDelInternal
% --------------------------------------------------------------------------
% deletes all internal solution sets and associated tables
%
solSetDelInternal () :-
    SDB=sdb::get(),
    SDB:chain_terms("H",HRec,_),
	HRec=cmp(_,[int(SolRef)|_]),
	SolRef<0,
	solSetDel(SolRef),
	fail
;
	!.	
% --------------------------------------------------------------------------
% solSetEnumerate
% --------------------------------------------------------------------------
class predicates
solSetEnum1 : (string,count,chainDB::bt_selector) procedure.
clauses
% enumerates a solution set
%
solSetEnumerate(SolRef) :-
	SSolRef=toString(SolRef),
	sx::concat(["S",SSolRef],Chain),
	SN=sdb::getIx(Chain),
	% set has been enumerated already - re-enumerate
	SDB=sdb::get(),
	SDB:bt_close(SN),
	SDB:bt_delete(Chain),
	%
	SDB:chain_first(Chain,_),
	% at least on solution in the set
	% create an index called Sn, where n is the solution ref.
	SDB:bt_create(Chain,SN2,16,4,0),
	sdb::dropIx(Chain),
	sdb::putIx(Chain,SN2),
	%	
	Counter=count::newCounter(0),
	solSetEnum1(Chain,Counter,SN),
	N=Counter:get,
	solSetSizeSet(SolRef,N),
			S1=toString(N),
			sx::concat(["Sol. set ",SSolRef," enumerated with ",S1," solutions."],Mess),
			tt::tt_ss(1,Mess),
			!
;
	SSolRef=toString(SolRef),
	sx::concat(["S",SSolRef],Chain),
	SDB=sdb::get(),
	SDB:chain_first(Chain,_),
	% at least on solution in the set
	% creates an index called Sn, where n is the solution ref.
	SDB:bt_create(Chain,SN,16,4,0),
	sdb::dropIx(Chain),
	sdb::putIx(Chain,SN),
	%	
	Counter=count::newCounter(0),
	solSetEnum1(Chain,Counter,SN),
	N=Counter:get,
	solSetSizeSet(SolRef,N),
			S1=toString(N),
			sx::concat(["Sol. set ",SSolRef," enumerated with ",S1," solutions."],Mess),
			tt::tt_ss(1,Mess),
			!
;
	% no solution in set	
	solSetSizeSet(SolRef,0),
			SSolRef=toString(SolRef),
			sx::concat(["Sol. set ",SSolRef," enumerated with 0 solutions."],Mess),
			tt::tt_ss(1,Mess),
			!.
% always succeeds
%	
solSetEnum1(Chain,Counter,SN) :-
	SDB=sdb::get(),
	SDB:chain_terms(Chain,SRec,Ref),
	Counter:inc(Next),
	SRec=cmp(_,[_|Vals]),
	NRec=cmp("S",[int(Next)|Vals]),
	SDB:term_replace(Ref,NRec),
	Key=toString(Next),
	SDB:key_insert(SN,Key,Ref),
	fail
;
	!.
/*	not needed 27.5.5
% --------------------------------------------------------------------------
% solSetExists
% --------------------------------------------------------------------------
% succeeds only if given query exists in a solution set and the solution  set is not altered
%
solSetExists(Query,SolRef) :-
    SDB=sdb::get(),
	SDB:chain_terms("H",HRec,_),
	HRec=cmp(_,[int(SolRef),_,_,_,Query,int(0)]),
	!.
*/	
% --------------------------------------------------------------------------
% solSetHeader
% --------------------------------------------------------------------------
% returns the solution set header record
%
solSetHeader(SolRef,NSols,Traced,NComps,Query,Altered) :-
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[_,int(NSols),int(Traced),int(NComps),Query,int(Altered)]),!
;
	dlgIF::feil("solSetHeader",[],[SolRef],nill),!,fail.	
% --------------------------------------------------------------------------
% solSetMarkAltered
% --------------------------------------------------------------------------
solSetMarkAltered(SolRef) :-
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[R,N,_,C,Q,_]),
	NRec=cmp("H",[R,N,int(0),C,Q,int(1)]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("solSetMarkAltered",[],[SolRef],nill),!,fail.						
% --------------------------------------------------------------------------
% solSetNCompsSet
% --------------------------------------------------------------------------
solSetNCompsSet(SolRef,NComps) :- 
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[R,N,T,_,Q,A]),
	NRec=cmp("H",[R,N,T,int(NComps),Q,A]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("solSetNCompsSet",[],[SolRef,NComps],nill),!,fail.					
% --------------------------------------------------------------------------
% solSetNew
% --------------------------------------------------------------------------
class predicates
solSetCounter : (integer) procedure.
solSetNewX : (integer) procedure.
clauses
% initializes a new solution set
%
solSetNew(Query,SolRef) :-
	N=globalVars::get_i("piercl"),
	N>0,
	globalVars::decr("solint",_,SolRef),
	solSetCounter(N),
	solSetNewX(SolRef),
	HRec=cmp("H",[int(SolRef),int(0),int(0),int(0),Query,int(0)]),
	SDB=sdb::get(),
	SDB:chain_inserta("H",HRec,Ref),
	Key1=toString(SolRef),
	Ix1=sdb::getIx("H1"),
	SDB:key_insert(Ix1,Key1,Ref),
			tt::tt_s(2,"Initialized PIEX solution set",Key1),
	!
;
    globalVars::get_i("piercl")=0,
	globalVars::incr("solref",_,SolRef),
	solSetCounter(0),
	solSetNewX(SolRef),
	HRec=cmp("H",[int(SolRef),int(0),int(0),int(0),Query,int(0)]),
	SDB=sdb::get(),
	SDB:chain_inserta("H",HRec,Ref),
	Key1=toString(SolRef),
	Ix1=sdb::getIx("H1"),
	SDB:key_insert(Ix1,Key1,Ref),
			tt::tt_s(2,"Initialized PIEX solution set",Key1),
	!
;
	dlgIF::feil("solSetNew",[],[],Query),!,fail.	
solSetCounter(0) :- 
	globalVars::set_i("solcnt0",0),	
	globalVars::set_i("solcnt1",0),	
	globalVars::set_i("solcnt2",0),	
	globalVars::set_i("solcnt3",0),	
	globalVars::set_i("solcnt4",0),!.	
solSetCounter(1) :- 
	globalVars::set_i("solcnt1",0),	
	globalVars::set_i("solcnt2",0),	
	globalVars::set_i("solcnt3",0),	
	globalVars::set_i("solcnt4",0),!.	
solSetCounter(2) :- 
	globalVars::set_i("solcnt2",0),	
	globalVars::set_i("solcnt3",0),	
	globalVars::set_i("solcnt4",0),!.	
solSetCounter(3) :- 
	globalVars::set_i("solcnt3",0),	
	globalVars::set_i("solcnt4",0),!.	
solSetCounter(4) :- 
	globalVars::set_i("solcnt4",0),!.	
solSetCounter(_) :- !.
solSetNewX(SolRef) :-
	N=SolRef mod 50,
	N=0,
	sdb::sdbSetStatus,
	!
;
	!.		
% --------------------------------------------------------------------------
% solSetSize
% --------------------------------------------------------------------------
solSetSize(SolRef,N) :- 
    solSetHeader(SolRef,N,_,_,_,_),
    !.
% --------------------------------------------------------------------------
% solSetSizeSet
% --------------------------------------------------------------------------
solSetSizeSet(SolRef,N) :- 
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[R,_,T,C,Q,A]),
	NRec=cmp("H",[R,int(N),T,C,Q,A]),
	SDB:term_replace(Ref,NRec),
	!
;
	dlgIF::feil("solSetSizeSet",[],[SolRef],nill),!,fail.					
% --------------------------------------------------------------------------
% solSetToSentences
% --------------------------------------------------------------------------
class predicates
solConstrSent : (sterm,string_list,integer,integer,sterm) determ (i,i,i,i,o).
solSetToSent : (integer,integer,integer,sterm,string_list,sterml) determ (i,i,i,i,i,o).
clauses
solSetToSentenceTerms(ST,SolRef,Sentences) :-
	solST_Var(ST,STv,Vs),
	transpr::proPR_Term(STv,STt),
	solSetSize(SolRef,N),
	solSetToSent(SolRef,1,N,STt,Vs,Sentences),
	!.
	
solSetToSent(_,SolNo,SolMax,_,_,[]) :-
	SolNo>SolMax,
	!.
solSetToSent(SolRef,SolNo,SolMax,STt,Vs,[Sentence|Sentences]) :-
	solConstrSent(STt,Vs,SolRef,SolNo,Sentence),
	Next=SolNo+1,
	solSetToSent(SolRef,Next,SolMax,STt,Vs,Sentences),
	!.
solConstrSent(STt,[],_,_,STt) :-
	!.
solConstrSent(STt,[V|Vs],SolRef,SolNo,STd) :-
	solCompVar(SolRef,SolNo,V,Val),
	struc::treReplace(STt,var(V),Val,STm),
	solConstrSent(STm,Vs,SolRef,SolNo,STd),
	!.
solST_Var(STk,STv,Vs) :-
	sx::sxUnpack(STk,'|',L),
	L=[P,Ea,Eb],
	dbBase::smClassDef(P,"p",_,PDEs,_,_,N),
	N>0,
	dbBase::smClassDef(Ea,"e",_,EaDEs,_,_,_),
	dbBase::smClassDef(Eb,"e",_,EbDEs,_,_,_),
	vars::varGenDE_VarsSNo(P,0,PDEs,PVs),
	vars::varGenDE_VarsSNo(Ea,1,EaDEs,EaVs),
	vars::varGenDE_VarsSNo(Eb,2,EbDEs,EbVs),
	set::conc([EaVs,EbVs,PVs],Vs),
	sx::sxlJoin(PVs,",",SP),
	sx::sxlJoin(EaVs,",",SEa),
	sx::sxlJoin(EbVs,",",SEb),
	sx::concat([P,"(",Ea,"(",SEa,"),",Eb,"(",SEb,"),",SP,")"],STv),
	!
;	
	sx::sxUnpack(STk,'|',L),
	L=[P,Ea,Eb],
	dbBase::smClassDef(Ea,"e",_,EaDEs,_,_,_),
	dbBase::smClassDef(Eb,"e",_,EbDEs,_,_,_),
	vars::varGenDE_VarsSNo(Ea,1,EaDEs,EaVs),
	vars::varGenDE_VarsSNo(Eb,2,EbDEs,EbVs),
	set::conc([EaVs,EbVs],Vs),
	sx::sxlJoin(EaVs,",",SEa),
	sx::sxlJoin(EbVs,",",SEb),
	sx::concat([P,"(",Ea,"(",SEa,"),",Eb,"(",SEb,"))"],STv),
	!
;	
	sx::sxUnpack(STk,'|',L),
	L=[P,Ea],
	dbBase::smClassDef(P,"p",_,PDEs,_,_,N),
	N>0,
	dbBase::smClassDef(Ea,"e",_,EaDEs,_,_,_),
	vars::varGenDE_VarsSNo(P,0,PDEs,PVs),
	vars::varGenDE_VarsSNo(Ea,1,EaDEs,EaVs),
	set::conc([EaVs,PVs],Vs),
	sx::sxlJoin(PVs,",",SP),
	sx::sxlJoin(EaVs,",",SEa),
	sx::concat([P,"(",Ea,"(",SEa,"),",SP,")"],STv),
	!
;	
	sx::sxUnpack(STk,'|',L),
	L=[P,Ea],
	dbBase::smClassDef(Ea,"e",_,EaDEs,_,_,_),
	vars::varGenDE_VarsSNo(Ea,1,EaDEs,Vs),
	sx::sxlJoin(Vs,",",SEa),
	sx::concat([P,"(",Ea,"(",SEa,"))"],STv),
	!
;
	dlgIF::feil("solST_Var",[STk],[],nill),!,fail.
% --------------------------------------------------------------------------
% solSetTracedSet
% --------------------------------------------------------------------------
% sets the traced-flag for the solution set
%
solSetTracedSet(SolRef) :-
	H1=sdb::getIx("H1"),
	Key=toString(SolRef),
	SDB=sdb::get(),
	SDB:key_search(H1,Key,Ref),
	SDB:ref_term(Ref,HRec),
	HRec=cmp(_,[R,N,int(0)|T]),
	NRec=cmp("H",[R,N,int(1)|T]),
	SDB:term_replace(Ref,NRec),
	!
;
	!.	
% --------------------------------------------------------------------------
% solSwap
% --------------------------------------------------------------------------
solSwap(SolRef,N1,N2) :-
	S1=toString(SolRef),
	Ix=string::concat("S",S1),
	SN=sdb::getIx(Ix),
	Key1=toString(N1),
	SDB=sdb::get(),
	SDB:key_search(SN,Key1,Ref1),
	SDB:ref_term(Ref1,SRec1),
	SRec1=cmp(_,[_|T1]),
	NRec1=cmp("S",[int(N2)|T1]),
	%
	Key2=toString(N2),
	SDB:key_search(SN,Key2,Ref2),
	SDB:ref_term(Ref2,SRec2),
	SRec2=cmp(_,[_|T2]),
	NRec2=cmp("S",[int(N1)|T2]),
	%
	SDB:key_delete(SN,Key1,Ref1),
	SDB:key_delete(SN,Key2,Ref2),
	SDB:term_replace(Ref1,NRec1),
	SDB:term_replace(Ref2,NRec2),
	SDB:key_insert(SN,Key2,Ref1),
	SDB:key_insert(SN,Key1,Ref2),!
;
	dlgIF::feil("solSwap",[],[SolRef,N1,N2],nill),!,fail.	
% --------------------------------------------------------------------------
% solVarNo/s
% --------------------------------------------------------------------------
% returns the solution component no of variable
%
solVarNo(SolRef,Var,CompNo) :-
	V2=sdb::getIx("V2"),
	SSolRef=toString(SolRef),
	sx::concat([SSolRef,"|",Var],Key),
	SDB=sdb::get(),
	SDB:key_search(V2,Key,Ref),
	SDB:ref_term(Ref,VRec),
	VRec=cmp(_,[_,int(CompNo),_]),!
;
	dlgIF::feil("solVarNo",[Var],[SolRef],nill),!,fail.	
solVarNos(_,[],[]) :-
	!.
solVarNos(SolRef,[Var|Vars],[Comp|Comps]) :-
	solVarNo(SolRef,Var,Comp),
	solVarNos(SolRef,Vars,Comps),
	!.
% --------------------------------------------------------------------------
% solSort/V
% --------------------------------------------------------------------------
class predicates
solSort1 : (integer,integer,integer_list) determ.
solSortSet : (integer,integer,integer_list) determ.
clauses
%
% a solution set is not sorted directly, it is sorted by the generation of a table
% corresponding to the solution  set. the table is then sorted, and the solution  set is 
% thereafter accessed in sorted order through this table. 
%
% sorting of sets are only used within the inf. engine when evaluating procedure rules.
% parameters:
% 1	in	solution ref.
% 2	in	list of solution  component nos corresponding to the sort keys.
%    		18.6.00 no longer supported:
%		this list can be empty in which case a dummy sorting is carried out based\on component#1
%
solSort(SolRef,[]) :- 
	tt::tt_i(1,"Empty sorting key - set not sorted",SolRef),
	!.
solSort(SolRef,Keys) :-
	solSetSize(SolRef,N),
	N=0,
	dlgIF::feil("solSort/1",[],[SolRef|Keys],nill),!,fail
;
	stb::stbM_GetF(SolRef,_,_,_,_,_), 
	% this is a re-sort
	stb::stbDeleteSolAll(SolRef),
	solSort(SolRef,Keys),
	!
;
	% tables are not generated 
	solSetHeader(SolRef,_,_,_,Query,_),
	table::tabGen(0,SolRef,Query),
			tt::tt_table(4,"sort table (unsorted)",SolRef),
	stb::stbM_GetF(SolRef,NTables,_,_,_,_),
	solSort1(SolRef,NTables,Keys),
	!
;
	dlgIF::feil("solSort",[],[SolRef|Keys],nill),!,fail.	
solSort1(SolRef,1,Keys) :-
	solSortSet(SolRef,1,Keys),
			tt::tt_is(4,"sort-keys",Keys),
	stb::stbSort(SolRef,1),
	!.
solSort1(SolRef,N,Keys) :- 
	dlgIF::feil("solSort1",[],[SolRef,N|Keys],nill),!,fail.
solSortSet(_,_,[]) :- 
	!.
solSortSet(SolRef,N,[Key|Keys]) :-
	stb::stbD_SetSort(SolRef,1,Key,N),
	Next=N+1,
	solSortSet(SolRef,Next,Keys),
	!.
% ----------------------------------------------------------------------------------------------
% tt_sors
% ----------------------------------------------------------------------------------------------
class predicates
tt_Sols : (integer,integer,string,integer).
tt_SolComps : (integer,integer,integer,integer) determ.
clauses

tt_sors(Lev,S,SolRef) :-
	Lev<0,
	L=-Lev,
	tt_Sols(L,1,S,SolRef),!
;
	tt_Sols(Lev,0,S,SolRef),!.
		
tt_Sols(Lev,Always,S,SolRef) :-
	tt::getLevel()<Lev, 
	!
;	
	Always=0,
	sol::solSetHeader(SolRef,NSol,Traced,_,_,_),
	Traced=1,
	S1=toString(SolRef),
	S3=toString(NSol),
	sx::concat(["SOLUTION SET ",S1," (",S3,") SEE ABOVE"],S4),
	tt::ttHeader(Lev,S4,S),
	tt::wrl([]),
	tt::tt_cc,!
;
	sol::solSetSize(SolRef,N),
	N=0,
	S1=toString(SolRef),
	sx::concat(["SOLUTION SET ",S1," (0) IS EMPTY"],S4),
	tt::ttHeader(Lev,S4,S),
	tt::wrl([]),
	tt::tt_cc,!
;
	sol::solSetHeader(SolRef,NSol,_,NComps,_,_),
	S1=toString(SolRef),
	S3=toString(NSol),
	sx::concat(["SOLUTION SET ",S1," (#=",S3,")"],S4),
	tt::ttHeader(Lev,S4,S),tt::wrl([]),
	sol::solComps(SolRef,_,Comps),
	tt::tt_strings(Lev,"solution components",Comps),
	tt::tt_co,
	tt::wrl(["SOLUTIONS"]),
	sol::forallsol(SolRef,SolNo),
	   tt::tt_co,
	   tt_SolComps(SolRef,SolNo,1,NComps),tt::wrl([]),
	   fail
;	
	sol::solSetTracedSet(SolRef),
	tt::tt_cc,!.

tt_SolComps(_,_,N,M) :- N>M,!.	

tt_SolComps(SolRef,SolNo,CompNo,M) :-
	sol::solCompNo(SolRef,SolNo,CompNo,Val),
	SSolRef=toString(SolRef),
	SSolNo=toString(SolNo),
	SCompNo=toString(CompNo),
	SVal=toString(Val),
	tt::wrl(["     Set=",SSolRef," #",SSolNo,".",SCompNo," = ",SVal]),
	Next=CompNo+1,
	tt_SolComps(SolRef,SolNo,Next,M),!	
;
	dlgIF::feil("tt_SolComps",[],[SolRef,SolNo,CompNo],nill),!,fail.	
end implement sol
